#include <math.h>
#include <cublas_v2.h>
#include <cufft.h>
#include "sb_macro.h"
#include "sb_common.h"
#include "solve_poisson_2d.h"
#if MATVECWITHCUSPARSE
	#include <cusparse_v2.h>
#endif

double *SBFL( int N, int P, double *dataX, double *dataY, double lamb1, double lamb2, double mu1, double mu2, double epsil, double tol, double *obj, int *niter)
{
	// Host memory
	double alpha, obj_old, obj_new, max_diff, normb, error, toltol, gamma0, gamma, tgamma;
	double *result;
	size_t sizeN = N * sizeof(double);
	size_t sizeP = P * sizeof(double);
	size_t sizeP1 = ( P - 1 ) * sizeof(double);
	int i, k, iter, nb = (int)( ((double) P / THREADS ) + 0.999 ), M = sqrt(P);
	double mu11 = 1 / mu1,			mu22 = 1 / mu2;
	double smu11 = 1 / sqrt(mu1),	smu22 = 1 / sqrt(mu2);
	double smu1 = sqrt(mu1),		smu2 = sqrt(mu2);
	double nmu1 = -mu1,				nmu2 = -mu2;
	double *tttmp = (double *) malloc (sizeP);
	FILE *f;

	// Device memory
	double *beta, *r, *XTY, *a, *w, *amw, *z, *t, *p, *Ax, *Lbv, *x, *q; // sizeP
	double *B, *Ap, *tvP1, *tvP2, *tvP3, *tvP1_d2; // _d2 is for device 2
	double *beta1, *b, *v, *bv; // sizeP1
	double *Y, *resid; // sizeN

	// Device setting
	int deviceCount;
	int deviceList[10];
	struct cudaDeviceProp prop;
	cudaGetDeviceCount( &deviceCount );
	k = 0;
	for (i=0; i<deviceCount; i++)
	{
		cudaGetDeviceProperties( &prop, i );
		if ( prop.totalGlobalMem > 1024 * 1024 * 1024 ) // Choosing a device whose global memory > 1GB
			deviceList[k++] = i;
	}
	deviceCount = k;

	if ( ( 8.0 * N * P / 1024 / 1024 / 1024 ) > ( 0.9 * GPUGLOBALMEMORY ) )
	{
		printf("Not enough memory. ");
		if ( deviceCount > 1 )
			printf("Running in multi-GPU mode (%d GPUs found).\n", deviceCount);
		else
			printf("It may cause segmentation faults.\n");
	}
	cudaDeviceReset();
 
	cublasHandle_t handle;
	cublasCreate( &handle );
	
	#if MULTIGPU
		int tmpN = (int)( N / deviceCount );

		double *Xblock[deviceCount];
		double *Yblock[deviceCount], *residblock[deviceCount];
		double *sizePblock[deviceCount], *sizePblock_d0[deviceCount], *sizePblock_d2[deviceCount];
		cudaEvent_t stop;
		cudaEventCreate( &stop );
		cudaStream_t stream[deviceCount];
		cublasHandle_t handles[deviceCount];
		for (i=0; i<deviceCount; i++)
		{
			cudaSetDevice( deviceList[i] );
			myMalloc( Xblock[i], tmpN * P * sizeof(double) );
			myMalloc( Yblock[i], tmpN * sizeof(double) );
			myMalloc( residblock[i], tmpN * sizeof(double) );
			myMalloc( sizePblock[i], sizeP );
			cudaStreamCreate( &(stream[i]) );
			cublasCreate( &(handles[i]) );
			cublasSetStream( handles[i], stream[i] );
		}
		cudaSetDevice(0);
		for (i=0; i<2; i++)
			myMalloc( sizePblock_d0[i], sizeP );
		cudaSetDevice(2);
		for (i=0; i<2; i++)
			myMalloc( sizePblock_d2[i], sizeP );
		cudaSetDevice(0);
		for (i=0; i<deviceCount; i++)
		{
			cudaMemcpy( Xblock[i], dataX + i*tmpN*P, tmpN * P * sizeof(double), HtoD );
			cudaMemcpy( Yblock[i], dataY + i*tmpN, tmpN * sizeof(double), HtoD );
		}
		cudaEventRecord(stop);
		cudaEventSynchronize(stop);
	#else
		bool PreferXTX = ( N >= P ) && ( N > 256 ) && ( P < 6120 );
		size_t sizeNP = N * P * sizeof(double);

		double *X, *XTX;
		myMalloc( X, sizeNP );
		cudaMemcpy( X, dataX, sizeNP, HtoD );
		if ( PreferXTX )
		{
			myMalloc( XTX, P * P * sizeof(double) );
			cudaMemset( (void *)XTX, 0, P * P * sizeof(double) );
			cublasDsyrk( handle, CUBLAS_FILL_MODE_UPPER, CUBLAS_OP_N, P, N, &ONE, X, P, &ZERO, XTX, P );
			cublasDsyrk( handle, CUBLAS_FILL_MODE_LOWER, CUBLAS_OP_N, P, N, &ONE, X, P, &ZERO, XTX, P );
			//cublasDgemm( handle, CUBLAS_OP_N, CUBLAS_OP_T, P, P, N, &ONE, X, P, X, P, &ZERO, XTX, P );
		}
	#endif
	
	#if MATVECWITHCUSPARSE
		cusparseHandle_t handle_sparse;
		cusparseCreate(&handle_sparse);

		cusparseMatDescr_t descr = 0;
		cusparseCreateMatDescr(&descr);
		cusparseSetMatType( descr, CUSPARSE_MATRIX_TYPE_GENERAL );
		cusparseSetMatIndexBase( descr, CUSPARSE_INDEX_BASE_ZERO );

		int nnz = P + 4*M*(M-1);
		double *cooValA_h;
		int *cooRowIndA_h;
		int *cooColIndA_h;
		double *cooValA;
		int *cooRowIndA;
		int *cooColIndA;
		int *csrRowPtr;
		cooValA_h = (double *) malloc( nnz*sizeof(double) );
		cooRowIndA_h = (int *) malloc( nnz*sizeof(int) );
		cooColIndA_h = (int *) malloc( nnz*sizeof(int) );
		cudaMalloc( (void **) &cooValA, nnz*sizeof(double) );
		cudaMalloc( (void **) &cooRowIndA, nnz*sizeof(int) );
		cudaMalloc( (void **) &cooColIndA, nnz*sizeof(int) );
		cudaMalloc( (void **) &csrRowPtr, (P+1)*sizeof(int) );
		k = 0;
		for (i=0; i<P; i++) {
			if ( ((int) (i / M)) > 0 ) {
				cooValA_h[k] = -mu2;
				cooRowIndA_h[k] = i;
				cooColIndA_h[k] = i-M;
				k++;
			}
			if ( (i % M) > 0 ) {
				cooValA_h[k] = -mu2;
				cooRowIndA_h[k] = i;
				cooColIndA_h[k] = i-1;
				k++;
			}
			cooValA_h[k] = mu1+4*mu2;
			cooRowIndA_h[k] = i;
			cooColIndA_h[k] = i;
			k++;
			if ( (i % M) < (M-1) ) {
				cooValA_h[k] = -mu2;
				cooRowIndA_h[k] = i;
				cooColIndA_h[k] = i+1;
				k++;
			}
			if ( ((int) (i / M)) < (M-1) ) {
				cooValA_h[k] = -mu2;
				cooRowIndA_h[k] = i;
				cooColIndA_h[k] = i+M;
				k++;
			}
		}
		cudaMemcpy( cooValA, cooValA_h, nnz*sizeof(double), HtoD );
		cudaMemcpy( cooRowIndA, cooRowIndA_h, nnz*sizeof(int), HtoD );
		cudaMemcpy( cooColIndA, cooColIndA_h, nnz*sizeof(int), HtoD );
		cusparseXcoo2csr( handle_sparse, cooRowIndA, nnz, P, csrRowPtr, CUSPARSE_INDEX_BASE_ZERO ); 
		free(cooValA_h);
		free(cooRowIndA_h);
		free(cooColIndA_h);
	#endif

	myMalloc( beta, 13 * sizeP );cudaMemset( beta, 0, 5 * sizeP );
	r		= beta + P;
	XTY		= r + P;
	a		= XTY + P;
	w		= a + P;
	amw		= w + P;
	z		= amw + P;
	t		= z + P;
	p		= t + P;
	Ax		= p + P;
	Lbv		= Ax + P;
	x		= Lbv + P;
	q		= x + P;

	myMalloc( B, 2 * (sizeN + sizeP + sizeP1) + 3 * sizeP );
	Ap		= B + N + P + (P-1);
	tvP1	= Ap + N + P + (P-1);
	tvP2	= tvP1 + P;
	tvP3	= tvP2 + P;
	cudaSetDevice(2);
	myMalloc( tvP1_d2, sizeP );
	cudaSetDevice(0);

	myMalloc( beta1, 4 * sizeP1 );		cudaMemset( beta1, 0, 3 * sizeP1 );
	b		= beta1 + (P-1);
	v		= b + (P-1);
	bv		= v + (P-1);

	myMalloc( Y, 2 * sizeN );			cudaMemcpy( Y, dataY, sizeN, HtoD );
	resid	= Y + N;

	result = (double *) malloc( sizeP );

	// Initialize XTY
	#if MULTIGPU
		cudaMemcpyMulti2Multi( residblock, Yblock, tmpN * sizeof(double) );
		cublasMultiDgemv( handles, CUBLAS_OP_N, P, tmpN, ONE, Xblock, P, residblock, 1, ZERO, sizePblock, 1 );
		cudaMemcpyMulti2Multi( sizePblock_d0, sizePblock, sizeP );
		cublasMultiDaxpy( handle, P, ONE, sizePblock_d0, 1, XTY, 1 ); // XTY = XTY + sizePblock_d0[i]
	#else
		cublasDgemv( handle, CUBLAS_OP_N, P, N, &ONE, X, P, Y, 1, &ZERO, XTY, 1 );
	#endif

	// Initialize FFT variables
	cufftHandle plan;
	cufftDoubleComplex *r_complex_d;
	double *kx_d;	myMalloc( kx_d, M * sizeof(double) );
	double *kx = (double *) malloc( M * sizeof(double) );
	#if DCTBASE
		size_t size4P8M4 = ( 4 * P + 8 * M + 4 ) * sizeof(double);
		int FFTrowsize = 2 * M + 2;
		int FFTindex[4] = { 2 * M + 3, 4 * M + 3, 4 * P + 6 * M + 3, 4 * P + 8 * M + 3 };
		double scale = -1.0 / ( M + 1 ) / ( M + 1 );
		cufftPlan2d( &plan, FFTrowsize, FFTrowsize, CUFFT_D2Z );
		cufftDoubleReal *r_real_d;
		myMalloc( r_real_d,		FFTrowsize * FFTrowsize * sizeof(cufftDoubleReal) );
		myMalloc( r_complex_d,	FFTrowsize * FFTrowsize * sizeof(cufftDoubleComplex) );
		//for (i=0; i<M; i++) kx[i] = mu1 * (double)M * (double)M / 2.0 + mu2 * (double)2.0 * ( 1 - cos(M_PI*(i+1)/(M+1)) );
		//for (i=0; i<M; i++) kx[i] = ( mu1 / 2.0 + mu2 * (double)2.0 * ( 1 - cos(M_PI*(i+1)/(M+1)) ) ) / (double)M;
		for (i=0; i<M; i++) kx[i] = ( mu1 / 2.0 + mu2 * (double)2.0 * ( 1 - cos(M_PI*(i+1)/(M+1)) ) );
	#else
		cufftPlan2d( &plan, M, M, CUFFT_Z2Z );
		cufftDoubleComplex *r_real_d;
		myMalloc( r_real_d,		M * M * sizeof(cufftDoubleComplex) );
		myMalloc( r_complex_d,	M * M * sizeof(cufftDoubleComplex) );
		for (i=0; i<M; i++) kx[i] = ( mu1 / 2.0 + mu2 * (double)2.0 * ( 1 - cos(2*M_PI*i/M) ) ) / (double)M;
	#endif
	cudaMemcpy( kx_d, kx, M * sizeof(double), HtoD );
	free( kx );

	// getObjValue (resid = X * beta - Y)
	#if MULTIGPU
		cudaMemcpyMulti2Multi( residblock, Yblock, tmpN * sizeof(double) );
		cudaMemcpySingle2Multi( sizePblock, beta, sizeP );
		cublasMultiDgemv( handles, CUBLAS_OP_T, P, tmpN, MINUSONE, Xblock, P, sizePblock, 1, ONE, residblock, 1 );
		cudaMemcpyResidBlock;
	#else
		cudaMemcpy( resid, Y, sizeN, DtoD );
		cublasDgemv( handle, CUBLAS_OP_T, P, N, &MINUSONE, X, P, beta, 1, &ONE, resid, 1 );
	#endif
	obj_new = getObjValue( handle, resid, beta, beta1, lamb1, lamb2, N, P, sizeP1 );
printf("First obj-val: %f\n",obj_new);
	// Start Iteration
	for (iter=0; iter<*niter; iter++)
	{
		obj_old = obj_new;

		cudaMemcpy( bv, b, sizeP1, DtoD );						// bv = b
		cublasDscal( handle, P-1, &mu2, bv, 1 );				// bv = mu2*b
		cublasDaxpy( handle, P-1, &MINUSONE, v, 1, bv, 1 );		// bv = mu2*b - v
		cublasDscal( handle, P-1, &smu22, bv, 1 );				// bv = 1/sqrt(mu2)*bv

		cudaMemcpy( amw, a, sizeP, DtoD );						// amw = a
		cublasDscal( handle, P, &mu1, amw, 1 );					// amw = mu1*a
		cublasDaxpy( handle, P, &MINUSONE, w, 1, amw, 1 );		// amw = mu1*a - w
		cublasDscal( handle, P, &smu11, amw, 1 );				// amw = 1/sqrt(mu1)*amw
		
		cudaMemcpy( B, Y, sizeN, DtoD );						// B = [y; amw; bv]
		cudaMemcpy( B+N, amw, sizeP, DtoD );
		cudaMemcpy( B+N+P, bv, sizeP1, DtoD );
		//cublasDnrm2( handle, P, B, 1, &normb );

		cudaMemset( Lbv, 0, sizeof(double) );
		cudaMemcpy( Lbv+1, bv, sizeP1, DtoD );					// Lbv = [0 bv(1:end-1)]
		cublasDaxpy( handle, P-1, &MINUSONE, bv, 1, Lbv, 1 );	// Lbv = Lbv - bv

		cudaMemset( x, 0, sizeP );

		cudaMemcpy( q, XTY, sizeP, DtoD );						// q = X'*y
		cublasDaxpy( handle, P, &smu1, amw, 1, q, 1 );			// q = X'*y + (mu1*a - w)
		cublasDaxpy( handle, P, &smu2, Lbv, 1, q, 1 );			// q = X'*y + (mu1*a - w) + L'*(mu2*b - v)
		cublasDnrm2( handle, P, q, 1, &normb );					// normb = norm(q,2)

		#if DCTBASE
			DCTsolver_pcgls( z, q );							// z = linear system solution
		#else
			DFTsolver_pcgls( z, q );
		#endif
		//cublasDnrm2( handle, P, z, 1, &normb );					// normb = norm(z,2)
		cublasDnrm2( handle, P, z, 1, &gamma0 );

		gamma0 = gamma0 * gamma0;

		cudaMemcpy( p, z, sizeP, DtoD );

		toltol = tol;
		for (k=0; k<N; k++)
		{
			#if DCTBASE
				DCTsolver_pcgls( t, p );						// t = linear system solution
			#else
				DFTsolver_pcgls( t, p );
			#endif

			// Ap(1:N) = X * t
			#if MULTIGPU
				cudaMemcpySingle2Multi( sizePblock, t, sizeP );
				cublasMultiDgemv( handles, CUBLAS_OP_T, P, tmpN, ONE, Xblock, P, sizePblock, 1, ZERO, residblock, 1 );
				for (i=0; i<deviceCount; i++)
					cudaMemcpyAsync( Ap + i*tmpN, residblock[i], tmpN * sizeof(double), DtoD );
				cudaEventRecord(stop);
				cudaEventSynchronize(stop);
			#else
				cublasDgemv( handle, CUBLAS_OP_T, P, N, &ONE, X, P, t, 1, &ZERO, Ap, 1 );
			#endif

			// Ap(N+1:N+P) = smu1 * t
			cudaMemcpy( Ap+N, t, sizeP, DtoD );
			cublasDscal( handle, P, &smu1, Ap+N, 1 );

			// Ap(N+P+1:N+P+P-1) = smu2 * L * t
			cudaMemset( Ap+N+P+P-2, 0, sizeof(double) );
			cudaMemcpy( Ap+N+P, t+1, sizeP1, DtoD );
			cublasDaxpy( handle, P-1, &MINUSONE, t, 1, Ap+N+P, 1 );

			cublasDnrm2( handle, N+P+P-1, Ap, 1, &alpha );			// alpha = norm(Ap,2).^2
			alpha = alpha * alpha;
			alpha = gamma0 / alpha;
			cublasDaxpy( handle, P, &alpha, t, 1, x, 1 );			// x = x + alpha * t
			alpha = -alpha;
			cublasDaxpy( handle, P, &alpha, Ap, 1, B, 1 );			// B = B - alpha * Ap
			
			// q = A' * B
			#if MULTIGPU
				for (i=0; i<deviceCount; i++)
					cudaMemcpyAsync( residblock[i], Ap + i*tmpN, tmpN * sizeof(double), DtoD );
				cudaEventRecord(stop);
				cudaEventSynchronize(stop);
				cublasMultiDgemv( handles, CUBLAS_OP_N, P, tmpN, ONE, Xblock, P, residblock, 1, ZERO, sizePblock, 1 );
				cudaMemcpyAsync( sizePblock_d0[0], sizePblock[0], sizeP, DtoD );
				cudaMemcpyAsync( sizePblock_d0[1], sizePblock[1], sizeP, DtoD );
				cudaMemcpyAsync( sizePblock_d2[0], sizePblock[2], sizeP, DtoD );
				cudaMemcpyAsync( sizePblock_d2[1], sizePblock[3], sizeP, DtoD );
				cudaEventRecord(stop);
				cudaEventSynchronize(stop);
				cudaMemset( tvP1, 0, sizeP );
				cudaMemset( tvP1_d2, 0, sizeP );
				for (i=0; i<2; i++)
				{
					cublasDaxpy( handle, P, &ONE, sizePblock_d0[i], 1, tvP1, 1 );
					cublasDaxpy( handle, P, &ONE, sizePblock_d2[i], 1, tvP1_d2, 1 );
				}
				cudaMemcpy( tvP2, tvP1_d2, sizeP, DtoD );
				cublasDaxpy( handle, P, &ONE, tvP2, 1, tvP1, 1 );
			#else
				cublasDgemv( handle, CUBLAS_OP_N, P, N, &ONE, X, P, Ap, 1, &ZERO, tvP1, 1 );
			#endif

			cudaMemcpy( tvP2, Ap+N, sizeP, DtoD );

			cudaMemset( tvP3, 0, sizeof(double) );
			cudaMemcpy( tvP3+1, Ap+N+P, sizeP1, DtoD );					// Lbv = [0 bv(1:end-1)]
			cublasDaxpy( handle, P-1, &MINUSONE, Ap+N+P, 1, tvP3, 1 );	// Lbv = Lbv - bv

			cublasDscal( handle, P, &alpha, tvP1, 1 );
			cublasDscal( handle, P, &alpha, tvP2, 1 );
			cublasDscal( handle, P, &alpha, tvP3, 1 );
			cublasDscal( handle, P, &smu1, tvP2, 1 );
			cublasDscal( handle, P, &smu2, tvP3, 1 );

			cublasDaxpy( handle, P, &ONE, tvP1, 1, q, 1 );
			cublasDaxpy( handle, P, &ONE, tvP2, 1, q, 1 );
			cublasDaxpy( handle, P, &ONE, tvP3, 1, q, 1 );

			#if DCTBASE
				DCTsolver_pcgls( z, q );							// z = linear system solution
			#else
				DFTsolver_pcgls( z, q );
			#endif
			
			if (iter<0 && k<0){
				cudaMemcpy( tttmp, z, sizeP, DtoH );
				f = fopen("z_pcgls.txt", "wb");
				fwrite(tttmp, sizeof(double), P, f);
				fclose(f);
				cudaMemcpy( tttmp, q, sizeP, DtoH );
				f = fopen("r_pcgls.txt", "wb");
				fwrite(tttmp, sizeof(double), P, f);
				fclose(f);
			}
			
			cublasDnrm2( handle, P, z, 1, &gamma );					// gamma0 = norm(z,2).^2
			gamma = gamma * gamma;

			tgamma = gamma / gamma0;
			cublasDscal( handle, P, &tgamma, p, 1 );
			cublasDaxpy( handle, P, &ONE, z, 1, p, 1 );				// p = z + gamma / gamma0 * p;
			gamma0 = gamma;

			if (iter<1 && k<0)
			{
				cublasDnrm2( handle, P, B, 1, &error );
				printf("r%f ", error);
				cublasDnrm2( handle, P, q, 1, &error );
				printf("u%f ", error);
				printf("\n");
			}
		
			//cublasDnrm2( handle, P, z, 1, &error );
			//if ( error / normb < toltol ) { k++; break; }
			cublasDnrm2( handle, P, q, 1, &error );
			if ( error / normb < 1e-5 ) { k++; break; }

		}

		cudaMemcpy( beta, x, sizeP, DtoD );

		cudaMemcpy( a, w, sizeP, DtoD );							// a = w
		cublasDscal( handle, P, &mu11, a, 1 );
		cublasDaxpy( handle, P, &ONE, beta, 1, a, 1 );				// a = beta + 1/mu1*w
		SoftThr<<<nb,THREADS>>>( a, lamb1, mu1, P );
		
		cublasDaxpy( handle, P, &mu1, beta, 1, w, 1 );				// w = w + mu1*beta
		cublasDaxpy( handle, P, &nmu1, a, 1, w, 1 );				// w = w - mu1*a

		cudaMemcpy( beta1, beta+1, sizeP1, DtoD );
		cublasDaxpy( handle, P-1, &MINUSONE, beta, 1, beta1, 1 );	// beta1 = beta1 - beta
		
		cudaMemcpy( b, v, sizeP1, DtoD );							// b = v
		cublasDscal( handle, P-1, &mu22, b, 1 );
		cublasDaxpy( handle, P, &ONE, beta1, 1, b, 1 );				// b = beta1 + 1/mu2*v
		SoftThr<<<nb,THREADS>>>( b, lamb2, mu2, P-1 );
		
		cublasDaxpy( handle, P-1, &mu2, beta1, 1, v, 1 );			// v = v + mu2*beta1
		cublasDaxpy( handle, P-1, &nmu2, b, 1, v, 1 );				// v = v - mu2*b

		// getObjValue (resid = X * beta - Y)
		#if MULTIGPU
			cudaMemcpySingle2Multi( sizePblock, beta, sizeP );
			cudaMemcpyMulti2Multi( residblock, Yblock, tmpN * sizeof(double) );
			cublasMultiDgemv( handles, CUBLAS_OP_T, P, tmpN, MINUSONE, Xblock, P, sizePblock, 1, ONE, residblock, 1 );
			cudaMemcpyResidBlock;
		#else
			cudaMemcpy( resid, Y, sizeN, DtoD );
			cublasDgemv( handle, CUBLAS_OP_T, P, N, &MINUSONE, X, P, beta, 1, &ONE, resid, 1 );
		#endif
		obj_new = getObjValue( handle, resid, beta, beta1, lamb1, lamb2, N, P, sizeP1 );

		max_diff = ABS( (obj_new - obj_old) / obj_old );

		printf("%d:%d:%.9f\n", (iter+1), k, obj_new);

		if ( ( max_diff < 5e-4 ) && ( iter > 10 ) ) { iter++; break; }

	}
		
	if (iter<0){
		cudaMemcpy( tttmp, x, sizeP, DtoH );
		f = fopen("sol_pcgls.txt", "wb");
		fwrite(tttmp, sizeof(double), P, f);
		fclose(f);
	}
	*niter = iter;
	*obj = obj_new;
	cudaMemcpy( result, beta, sizeP, DtoH );

	cublasDestroy( handle );
	#if MATVECWITHCUSPARSE
		cusparseDestroyMatDescr( descr );
		cusparseDestroy( handle_sparse );
		cudaFree( cooValA );
		cudaFree( cooRowIndA );
		cudaFree( cooColIndA );
		cudaFree( csrRowPtr );
	#endif

	#if MULTIGPU
		for (i=0; i<deviceCount; i++)
		{
			cudaSetDevice( deviceList[i] );
			cudaFree( Xblock[i] );
			cudaFree( Yblock[i] );
			cudaFree( residblock[i] );
			cudaFree( sizePblock[i] );
			cudaStreamDestroy( stream[i] );
			//cublasDestroy( &(handle[i]) );
		}
		cudaSetDevice(0);
		for (i=0; i<2; i++)
			cudaFree( sizePblock_d0[i] );
		cudaSetDevice(2);
		for (i=0; i<2; i++)
			cudaFree( sizePblock_d2[i] );
		cudaSetDevice(0);
	#else
		cudaFree( X );
		if ( PreferXTX ) cudaFree( XTX );
	#endif

	cudaFree( beta );
	cudaFree( beta1 );
	cudaFree( Y );

	cudaFree( r_real_d );
	cudaFree( r_complex_d );
	cudaFree( kx_d );
free(tttmp);
	return result;

}

double *SBFLdenoising( int N, int P, double *dataY, double lamb1, double lamb2, double mu1, double mu2, double epsil, double tol, double *obj, int *niter)
{
	// Host memory
	double alpha, obj_old, obj_new, max_diff, normb, error, toltol, gamma0, gamma, tgamma;
	double *result;
	size_t sizeN = N * sizeof(double);
	size_t sizeP = P * sizeof(double);
	size_t sizeP1 = ( P - 1 ) * sizeof(double);
	int i, k, iter, nb = (int)( ((double) P / THREADS ) + 0.999 ), M = sqrt(P);
	double mu11 = 1 / mu1,			mu22 = 1 / mu2;
	double smu11 = 1 / sqrt(mu1),	smu22 = 1 / sqrt(mu2);
	double smu1 = sqrt(mu1),		smu2 = sqrt(mu2);
	double nmu1 = -mu1,				nmu2 = -mu2;
	double *tttmp = (double *) malloc (sizeP);
	FILE *f;

	// Device memory
	double *beta, *r, *XTY, *a, *w, *amw, *z, *t, *p, *Ax, *Lbv, *x, *q; // sizeP
	double *B, *Ap, *tvP1, *tvP2, *tvP3;
	double *beta1, *b, *v, *bv; // sizeP1
	double *Y, *resid; // sizeN

	// Device setting
	int deviceCount;
	int deviceList[10];
	struct cudaDeviceProp prop;
	cudaGetDeviceCount( &deviceCount );
	k = 0;
	for (i=0; i<deviceCount; i++)
	{
		cudaGetDeviceProperties( &prop, i );
		if ( prop.totalGlobalMem > 1024 * 1024 * 1024 ) // Choosing a device whose global memory > 1GB
			deviceList[k++] = i;
	}
	deviceCount = k;

	if ( ( N * P * 8.0 / 1024 / 1024 / 1024 ) > ( GPUGLOBALMEMORY * 0.9 ) )
	{
		printf("Not enough memory. ");
		if ( deviceCount > 1 )
			printf("Running in multi-GPU mode (%d GPUs found).\n", deviceCount);
		else
			printf("It may cause segmentation faults.\n");
	}
	
	cudaDeviceReset();
 
	cublasHandle_t handle;
	cublasCreate( &handle );
	
	size_t sizeNP = N * P * sizeof(double);

	#if MATVECWITHCUSPARSE
		cusparseHandle_t handle_sparse;
		cusparseCreate(&handle_sparse);

		cusparseMatDescr_t descr = 0;
		cusparseCreateMatDescr(&descr);
		cusparseSetMatType( descr, CUSPARSE_MATRIX_TYPE_GENERAL );
		cusparseSetMatIndexBase( descr, CUSPARSE_INDEX_BASE_ZERO );

		int nnz = P + 4*M*(M-1);
		double *cooValA_h;
		int *cooRowIndA_h;
		int *cooColIndA_h;
		double *cooValA;
		int *cooRowIndA;
		int *cooColIndA;
		int *csrRowPtr;
		cooValA_h = (double *) malloc( nnz*sizeof(double) );
		cooRowIndA_h = (int *) malloc( nnz*sizeof(int) );
		cooColIndA_h = (int *) malloc( nnz*sizeof(int) );
		cudaMalloc( (void **) &cooValA, nnz*sizeof(double) );
		cudaMalloc( (void **) &cooRowIndA, nnz*sizeof(int) );
		cudaMalloc( (void **) &cooColIndA, nnz*sizeof(int) );
		cudaMalloc( (void **) &csrRowPtr, (P+1)*sizeof(int) );
		k = 0;
		for (i=0; i<P; i++) {
			if ( ((int) (i / M)) > 0 ) {
				cooValA_h[k] = -mu2;
				cooRowIndA_h[k] = i;
				cooColIndA_h[k] = i-M;
				k++;
			}
			if ( (i % M) > 0 ) {
				cooValA_h[k] = -mu2;
				cooRowIndA_h[k] = i;
				cooColIndA_h[k] = i-1;
				k++;
			}
			cooValA_h[k] = mu1+4*mu2;
			cooRowIndA_h[k] = i;
			cooColIndA_h[k] = i;
			k++;
			if ( (i % M) < (M-1) ) {
				cooValA_h[k] = -mu2;
				cooRowIndA_h[k] = i;
				cooColIndA_h[k] = i+1;
				k++;
			}
			if ( ((int) (i / M)) < (M-1) ) {
				cooValA_h[k] = -mu2;
				cooRowIndA_h[k] = i;
				cooColIndA_h[k] = i+M;
				k++;
			}
		}
		cudaMemcpy( cooValA, cooValA_h, nnz*sizeof(double), HtoD );
		cudaMemcpy( cooRowIndA, cooRowIndA_h, nnz*sizeof(int), HtoD );
		cudaMemcpy( cooColIndA, cooColIndA_h, nnz*sizeof(int), HtoD );
		cusparseXcoo2csr( handle_sparse, cooRowIndA, nnz, P, csrRowPtr, CUSPARSE_INDEX_BASE_ZERO ); 
		free(cooValA_h);
		free(cooRowIndA_h);
		free(cooColIndA_h);
	#endif

	myMalloc( beta, 13 * sizeP );cudaMemset( beta, 0, 5 * sizeP );
	r		= beta + P;
	XTY		= r + P;
	a		= XTY + P;
	w		= a + P;
	amw		= w + P;
	z		= amw + P;
	t		= z + P;
	p		= t + P;
	Ax		= p + P;
	Lbv		= Ax + P;
	x		= Lbv + P;
	q		= x + P;

	myMalloc( B, 2 * (sizeN + sizeP + sizeP1) + 3 * sizeP );
	Ap		= B + N + P + (P-1);
	tvP1	= Ap + N + P + (P-1);
	tvP2	= tvP1 + P;
	tvP3	= tvP2 + P;

	myMalloc( beta1, 4 * sizeP1 );		cudaMemset( beta1, 0, 3 * sizeP1 );
	b		= beta1 + (P-1);
	v		= b + (P-1);
	bv		= v + (P-1);

	myMalloc( Y, 2 * sizeN );			cudaMemcpy( Y, dataY, sizeN, HtoD );
	resid	= Y + N;

	result = (double *) malloc( sizeP );

	// Initialize XTY
	cudaMemcpy( XTY, Y, sizeP, DtoD );

	// Initialize FFT variables
	cufftHandle plan;
	cufftDoubleComplex *r_complex_d;
	double *kx_d;	myMalloc( kx_d, M * sizeof(double) );
	double *kx = (double *) malloc( M * sizeof(double) );
	#if DCTBASE
		size_t size4P8M4 = ( 4 * P + 8 * M + 4 ) * sizeof(double);
		int FFTrowsize = 2 * M + 2;
		int FFTindex[4] = { 2 * M + 3, 4 * M + 3, 4 * P + 6 * M + 3, 4 * P + 8 * M + 3 };
		double scale = -1.0 / ( M + 1 ) / ( M + 1 );
		cufftPlan2d( &plan, FFTrowsize, FFTrowsize, CUFFT_D2Z );
		cufftDoubleReal *r_real_d;
		myMalloc( r_real_d,		FFTrowsize * FFTrowsize * sizeof(cufftDoubleReal) );
		myMalloc( r_complex_d,	FFTrowsize * FFTrowsize * sizeof(cufftDoubleComplex) );
		//for (i=0; i<M; i++) kx[i] = mu1 * (double)M * (double)M / 2.0 + mu2 * (double)2.0 * ( 1 - cos(M_PI*(i+1)/(M+1)) );
		//for (i=0; i<M; i++) kx[i] = ( mu1 / 2.0 + mu2 * (double)2.0 * ( 1 - cos(M_PI*(i+1)/(M+1)) ) ) / (double)M;
		for (i=0; i<M; i++) kx[i] = ( mu1 / 2.0 + mu2 * (double)2.0 * ( 1 - cos(M_PI*(i+1)/(M+1)) ) );
	#else
		cufftPlan2d( &plan, M, M, CUFFT_Z2Z );
		cufftDoubleComplex *r_real_d;
		myMalloc( r_real_d,		M * M * sizeof(cufftDoubleComplex) );
		myMalloc( r_complex_d,	M * M * sizeof(cufftDoubleComplex) );
		for (i=0; i<M; i++) kx[i] = ( mu1 / 2.0 + mu2 * (double)2.0 * ( 1 - cos(2*M_PI*i/M) ) ) / (double)M;
	#endif
	cudaMemcpy( kx_d, kx, M * sizeof(double), HtoD );
	free( kx );

	// getObjValue (resid = Y - X * beta)
	cudaMemcpy( resid, Y, sizeN, DtoD );
	cublasDaxpy( handle, N, &MINUSONE, beta, 1, resid, 1 );
	obj_new = getObjValue( handle, resid, beta, beta1, lamb1, lamb2, N, P, sizeP1 );

	// Start Iteration
	for (iter=0; iter<*niter; iter++)
	{
		obj_old = obj_new;

		cudaMemcpy( bv, b, sizeP1, DtoD );						// bv = b
		cublasDscal( handle, P-1, &mu2, bv, 1 );				// bv = mu2*b
		cublasDaxpy( handle, P-1, &MINUSONE, v, 1, bv, 1 );		// bv = mu2*b - v
		cublasDscal( handle, P-1, &smu22, bv, 1 );				// bv = 1/sqrt(mu2)*bv

		cudaMemcpy( amw, a, sizeP, DtoD );						// amw = a
		cublasDscal( handle, P, &mu1, amw, 1 );					// amw = mu1*a
		cublasDaxpy( handle, P, &MINUSONE, w, 1, amw, 1 );		// amw = mu1*a - w
		cublasDscal( handle, P, &smu11, amw, 1 );				// amw = 1/sqrt(mu1)*amw
		
		cudaMemcpy( B, Y, sizeN, DtoD );						// B = [y; amw; bv]
		cudaMemcpy( B+N, amw, sizeP, DtoD );
		cudaMemcpy( B+N+P, bv, sizeP1, DtoD );
		//cublasDnrm2( handle, P, B, 1, &normb );

		cudaMemset( Lbv, 0, sizeof(double) );
		cudaMemcpy( Lbv+1, bv, sizeP1, DtoD );					// Lbv = [0 bv(1:end-1)]
		cublasDaxpy( handle, P-1, &MINUSONE, bv, 1, Lbv, 1 );	// Lbv = Lbv - bv

		cudaMemset( x, 0, sizeP );

		cudaMemcpy( q, XTY, sizeP, DtoD );						// q = X'*y
		cublasDaxpy( handle, P, &smu1, amw, 1, q, 1 );			// q = X'*y + (mu1*a - w)
		cublasDaxpy( handle, P, &smu2, Lbv, 1, q, 1 );			// q = X'*y + (mu1*a - w) + L'*(mu2*b - v)
		cublasDnrm2( handle, P, q, 1, &normb );					// normb = norm(q,2)

		#if DCTBASE
			DCTsolver_pcgls( z, q );							// z = linear system solution
		#else
			DFTsolver_pcgls( z, q );
		#endif
		//cublasDnrm2( handle, P, z, 1, &normb );					// normb = norm(z,2)
		cublasDnrm2( handle, P, z, 1, &gamma0 );

		gamma0 = gamma0 * gamma0;

		cudaMemcpy( p, z, sizeP, DtoD );

		toltol = tol;
		for (k=0; k<N; k++)
		{
			#if DCTBASE
				DCTsolver_pcgls( t, p );						// t = linear system solution
			#else
				DFTsolver_pcgls( t, p );
			#endif

			// Ap(1:N) = X * t
			cudaMemcpy( Ap, t, sizeP, DtoD );

			// Ap(N+1:N+P) = smu1 * t
			cudaMemcpy( Ap+N, t, sizeP, DtoD );
			cublasDscal( handle, P, &smu1, Ap+N, 1 );

			// Ap(N+P+1:N+P+P-1) = smu2 * L * t
			cudaMemset( Ap+N+P+P-2, 0, sizeof(double) );
			cudaMemcpy( Ap+N+P, t+1, sizeP1, DtoD );
			cublasDaxpy( handle, P-1, &MINUSONE, t, 1, Ap+N+P, 1 );

			cublasDnrm2( handle, N+P+P-1, Ap, 1, &alpha );			// alpha = norm(Ap,2).^2
			alpha = alpha * alpha;
			alpha = gamma0 / alpha;
			cublasDaxpy( handle, P, &alpha, t, 1, x, 1 );			// x = x + alpha * t
			alpha = -alpha;
			cublasDaxpy( handle, P, &alpha, Ap, 1, B, 1 );			// B = B - alpha * Ap
			
			// q = A' * B
			cudaMemcpy( tvP1, Ap, sizeP, DtoD );

			cudaMemcpy( tvP2, Ap+N, sizeP, DtoD );

			cudaMemset( tvP3, 0, sizeof(double) );
			cudaMemcpy( tvP3+1, Ap+N+P, sizeP1, DtoD );					// Lbv = [0 bv(1:end-1)]
			cublasDaxpy( handle, P-1, &MINUSONE, Ap+N+P, 1, tvP3, 1 );	// Lbv = Lbv - bv

			cublasDscal( handle, P, &alpha, tvP1, 1 );
			cublasDscal( handle, P, &alpha, tvP2, 1 );
			cublasDscal( handle, P, &alpha, tvP3, 1 );
			cublasDscal( handle, P, &smu1, tvP2, 1 );
			cublasDscal( handle, P, &smu2, tvP3, 1 );

			cublasDaxpy( handle, P, &ONE, tvP1, 1, q, 1 );
			cublasDaxpy( handle, P, &ONE, tvP2, 1, q, 1 );
			cublasDaxpy( handle, P, &ONE, tvP3, 1, q, 1 );

			#if DCTBASE
				DCTsolver_pcgls( z, q );							// z = linear system solution
			#else
				DFTsolver_pcgls( z, q );
			#endif
			
			if (iter<0 && k<0){
				cudaMemcpy( tttmp, z, sizeP, DtoH );
				f = fopen("z_pcgls.txt", "wb");
				fwrite(tttmp, sizeof(double), P, f);
				fclose(f);
				cudaMemcpy( tttmp, q, sizeP, DtoH );
				f = fopen("r_pcgls.txt", "wb");
				fwrite(tttmp, sizeof(double), P, f);
				fclose(f);
			}
			
			cublasDnrm2( handle, P, z, 1, &gamma );					// gamma0 = norm(z,2).^2
			gamma = gamma * gamma;

			tgamma = gamma / gamma0;
			cublasDscal( handle, P, &tgamma, p, 1 );
			cublasDaxpy( handle, P, &ONE, z, 1, p, 1 );				// p = z + gamma / gamma0 * p;
			gamma0 = gamma;

			if (iter<0 && k<0)
			{
				cublasDnrm2( handle, P, B, 1, &error );
				printf("r%f ", error);
				cublasDnrm2( handle, P, q, 1, &error );
				printf("u%f ", error);
				printf("\n");
			}
		
			//cublasDnrm2( handle, P, z, 1, &error );
			//if ( error / normb < toltol ) { k++; break; }
			cublasDnrm2( handle, P, q, 1, &error );
			if ( error / normb < tol ) { k++; break; }

		}
		
		cudaMemcpy( beta, x, sizeP, DtoD );

		cudaMemcpy( a, w, sizeP, DtoD );							// a = w
		cublasDscal( handle, P, &mu11, a, 1 );
		cublasDaxpy( handle, P, &ONE, beta, 1, a, 1 );				// a = beta + 1/mu1*w
		SoftThr<<<nb,THREADS>>>( a, lamb1, mu1, P );
		
		cublasDaxpy( handle, P, &mu1, beta, 1, w, 1 );				// w = w + mu1*beta
		cublasDaxpy( handle, P, &nmu1, a, 1, w, 1 );				// w = w - mu1*a

		cudaMemcpy( beta1, beta+1, sizeP1, DtoD );
		cublasDaxpy( handle, P-1, &MINUSONE, beta, 1, beta1, 1 );	// beta1 = beta1 - beta
		
		cudaMemcpy( b, v, sizeP1, DtoD );							// b = v
		cublasDscal( handle, P-1, &mu22, b, 1 );
		cublasDaxpy( handle, P, &ONE, beta1, 1, b, 1 );				// b = beta1 + 1/mu2*v
		SoftThr<<<nb,THREADS>>>( b, lamb2, mu2, P-1 );
		
		cublasDaxpy( handle, P-1, &mu2, beta1, 1, v, 1 );			// v = v + mu2*beta1
		cublasDaxpy( handle, P-1, &nmu2, b, 1, v, 1 );				// v = v - mu2*b

		// getObjValue (resid = X * beta - Y)
		cudaMemcpy( resid, Y, sizeN, DtoD );
		cublasDaxpy( handle, N, &MINUSONE, beta, 1, resid, 1 );
		obj_new = getObjValue( handle, resid, beta, beta1, lamb1, lamb2, N, P, sizeP1 );

		max_diff = ABS( (obj_new - obj_old) / obj_old );

		printf("%d:%d:%.9f\n", (iter+1), k, obj_new);

		if ( max_diff < epsil ) { iter++; break; }

	}

	*niter = iter;
	*obj = obj_new;
	cudaMemcpy( result, beta, sizeP, DtoH );

	cublasDestroy( handle );
	#if MATVECWITHCUSPARSE
		cusparseDestroyMatDescr( descr );
		cusparseDestroy( handle_sparse );
		cudaFree( cooValA );
		cudaFree( cooRowIndA );
		cudaFree( cooColIndA );
		cudaFree( csrRowPtr );
	#endif

	cudaFree( beta );
	cudaFree( beta1 );
	cudaFree( Y );

	cudaFree( r_real_d );
	cudaFree( r_complex_d );
	cudaFree( kx_d );
free(tttmp);
	return result;

}
